
import os
import sys
import json
import gzip

import pandas as pd
import numpy as np
from nameparser import HumanName
import unicodedata

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

from pyscisci.datasource.readwrite import load_preprocessed_data, load_int, load_float, load_bool
from pyscisci.database import BibDataBase

openalex_works_dfset = {'publications', 'references', 'publicationauthoraffiliation', 'concepts', 'fields', 'abstracts'}
openalex_dataframe_set = {'publications', 'works', 'venues', 'authors', 'institutions', 'affiliations', 'concepts', 'fields'}

class OpenAlex(BibDataBase):
    """
    Base class for the OpenAlex interface.

    This is an extension of 'BibDataBase' with processing functions specific to the MAG.
    See 'BibDataBase' in database.py for details of non-MAG specific functions.

    The MAG comes structured into three folders: mag, advanced, nlp.
    Explain downloading etc.

    """

    def __init__(self, path2database = '', database_extension='csv.gz', keep_in_memory = False, global_filter=None, 
        enable_dask=False, show_progress=True):

        self._default_init(path2database, database_extension, keep_in_memory, global_filter, enable_dask, show_progress)

        self.PublicationIdType = int
        self.AffiliationIdType = int
        self.AuthorIdType = int
        self.JournalIdType = int

    def preprocess(self, dataframe_list = None, show_progress=True):
        """
        Bulk preprocess the MAG raw data.

        Parameters
        ----------
        dataframe_list: list, default None
            The list of DataFrames to preprocess.  If None, all MAG DataFrames are preprocessed.
            
        show_progress: bool, default True
            Show progress with processing of the data.

        """
        if dataframe_list is None or 'all' in dataframe_list:
            dataframe_list = openalex_works_dfset.union(openalex_dataframe_set)

        pubdataframe_list = set(dataframe_list).intersection(openalex_works_dfset)

        if 'affiliations' in dataframe_list or 'institutions' in dataframe_list:
            self.parse_affiliations(preprocess = True, show_progress=show_progress)

        if 'authors' in dataframe_list:
            self.parse_authors(preprocess = True, show_progress=show_progress)

        if 'publications' in dataframe_list or 'works' in dataframe_list:
            self.parse_publications(preprocess = True, dataframe_list=pubdataframe_list, show_progress=show_progress)

        if 'concepts' in dataframe_list or 'fields' in dataframe_list:
            self.parse_concepts(preprocess=True, show_progress=show_progress)

        if 'venues' in dataframe_list or 'journals' in dataframe_list:
            self.parse_venues(preprocess=True, show_progress=show_progress)

    def download_from_source(self, aws_access_key_id='', aws_secret_access_key='', specific_update='', aws_bucket = 'openalex',
        dataframe_list = 'all', rewrite_existing = False, edit_works = True, show_progress=True):

        """
        Download the OpenAlex files from Amazon Web Services.
        The files will be saved to the path specified by `path2database` into RawXML.

        Parameters
        ----------
        source_url: str, default 'ftp.ncbi.nlm.nih.gov'
            The base url for the ftp server from which to download.
        
        aws_access_key_id: str
            The acess key for AWS (not required)

        aws_secret_access_key: str
            The secret acess key for AWS (not required)

        specific_update: str
            Download only a specific update date, specified by the date in Y-M-D format, 2022-01-01.  
            If empty the full data is downloaded.

        dataframe_list: list
            The data types to download and save from OpenAlex.
                'all'
                'affiliations'
                'authors'
                'publications'
                'references'
                'publicationauthoraffiliation'
                'fields'
                'abstracts'

        rewrite_existing: bool, default False
            If True, overwrite existing files or if False, only download any missing files.

        edit_works: bool, default True
            If True, edit the works to remove pieces of data.  If False, force keeping all entries from the works file.

        show_progress: bool, default True
            Show progress with processing of the data.
        
        # edited from https://stackoverflow.com/questions/31918960/boto3-to-download-all-files-from-a-s3-bucket
        """

        try:
            import boto3
            from botocore import UNSIGNED
            from botocore.config import Config
        except ImportError:
            raise ImportError("To download the OpenAlex data, you must install the aws S3 client and python package boto3.")

        if len(aws_access_key_id) > 0 and len(aws_secret_access_key) > 0:
            s3_client = boto3.client(aws_access_key_id=aws_access_key_id,
              aws_secret_access_key=aws_secret_access_key)
        else:
            s3_client = boto3.client('s3', config=Config(signature_version=UNSIGNED))


        dataframe_list = set(dataframe_list)

        if dataframe_list is None or 'all' in dataframe_list:
            dataframe_list = dataframe_list.union(openalex_works_dfset).union(openalex_dataframe_set)
        else:
            # some dataframes have different names from those in pyscisci
            if 'publications' in dataframe_list: dataframe_list.add('works')
            if 'journals' in dataframe_list: dataframe_list.add('venues')
            if 'fields' in dataframe_list: dataframe_list.add('concepts')
            if 'affiliations' in dataframe_list: dataframe_list.add('institutions')

        # see if we have to do any editing
        if edit_works and any(not pubsub in dataframe_list for pubsub in ['works', 'references', 'publicationauthoraffiliation', 'concepts', 'abstracts']):
            edit_works = True
        else:
            edit_works = False
        

        
        if not rewrite_existing:
            files_already_downloaded = [os.path.join(dirpath.replace(self.path2database, ''), file) for (dirpath, dirnames, filenames) in os.walk(self.path2database) for file in filenames]
        else:
            files_already_downloaded = []

        keys = []
        dirs = []
        next_token = ''
        
        base_kwargs = {'Bucket':aws_bucket}
        while next_token is not None:
            kwargs = base_kwargs.copy()
            if next_token != '':
                kwargs.update({'ContinuationToken': next_token})
            results = s3_client.list_objects_v2(**kwargs)
            
            for cdict in results.get('Contents', {}):
                k = cdict.get('Key')

                if not k in files_already_downloaded:
                    if 'data' in k: 
                        if k.split('/')[1] in dataframe_list and (specific_update=='' or k.split('/')[2] == 'updated_date={}'.format(specific_update)):
                            if k[-1] != '/':
                                keys.append(k)
                            else:
                                dirs.append(k)
                    else:
                        keys.append(k)
            next_token = results.get('NextContinuationToken')

        # make all of the directories
        for d in dirs:
            dest_pathname = os.path.join(self.path2database, d)
            if not os.path.exists(os.path.dirname(dest_pathname)):
                os.makedirs(os.path.dirname(dest_pathname))
        
        # now download all of the 
        for k in tqdm(keys, desc='Downloading OpenAlex File Number', leave=True, disable=not show_progress):
            dest_pathname = os.path.join(self.path2database, k)

            if not os.path.exists(os.path.dirname(dest_pathname)):
                os.makedirs(os.path.dirname(dest_pathname))
            
            s3_client.download_file(aws_bucket, k, dest_pathname)

            # the works contains references, concepts, and astracts, check if we keep these
            if edit_works and (len(k.split('/')) > 1 and k.split('/')[1] == 'works' and k[-3:]=='.gz'):
                with gzip.open(dest_pathname, 'r') as infile:
                    with gzip.open(dest_pathname.split('.')[0] + '_temp.gz', 'w') as outfile:
                        for line in infile:
                            if line != '\n'.encode():
                                jline = json.loads(line)

                                if ( not ('fields' in dataframe_list or 'concepts' in dataframe_list) ) and 'concepts' in jline:
                                    del jline['concepts']

                                if ( not ('references' in dataframe_list) ) and 'referenced_works' in jline:
                                    del jline['referenced_works']

                                if ( not ('publicationauthoraffiliation' in dataframe_list) ) and 'authorships' in jline:
                                    del jline['authorships']

                                if ( not ('abstracts' in dataframe_list) ) and 'abstract_inverted_index' in jline:
                                    del jline['abstract_inverted_index']

                                newline = json.dumps(jline) + "\n"
                                outfile.write(newline.encode('utf-8'))

                if os.path.exists(dest_pathname):
                    os.remove(dest_pathname)
                os.rename(dest_pathname.split('.')[0] + '_temp.gz', dest_pathname)



    def clean_openalex_ids(self, oid):
        try:
            oid = oid.replace('https://openalex.org/', '')
            return int(oid[1:])
        except:
            return None

    def parse_affiliations(self, preprocess=True, specific_update='', show_progress=True):
        """
        Parse the OpenAlex Affilation raw data.

        Parameters
        ----------
        preprocess: bool, default True
            Save the processed data in new DataFrames.

        specific_update: str
            Parse only a specific update date, specified by the date in Y-M-D format, 2022-01-01.  
            If empty the full data is parsed.

        show_progress: bool, default True
            Show progress with processing of the data.


        Returns
        ----------
        DataFrame
            Affiliation DataFrame.
        """

        affil_column_names = ['AffiliationId', 'FullName', 'GridId', 'WikiPage', 'Country', 'City', 'Region', 'Latitude', 'Longitude', 'NumberPublications', 'NumberCitations']

        if specific_update == '' or specific_update is None:
            institution_dir = os.path.join(self.path2database, 'data', 'institutions')
        else:
            institution_dir = os.path.join(self.path2database, 'data/institutions', 'updated_date={}'.format(specific_update))

        if not os.path.exists(institution_dir):
            raise ValueError("The institutions data was not found in the DataBase path.  Please download the data before parsing.")
        else:
            files_to_parse = [os.path.join(dirpath, file) for (dirpath, dirnames, filenames) in os.walk(institution_dir) for file in filenames if '.gz' in file]


        if not os.path.exists(os.path.join(self.path2database, self.path2affiliation)):
            os.mkdir(os.path.join(self.path2database, self.path2affiliation))

        ifile = 0
        affiliation_info = []
        for file_name in tqdm(files_to_parse, desc='Affiliations', leave=True, disable=not show_progress):
            with gzip.open(os.path.join(institution_dir, file_name), 'r') as infile:
                for line in infile:
                    if line != '\n'.encode():
                        aline = json.loads(line)
                        affline = [self.clean_openalex_ids(aline['id']), aline.get('display_name', None)]
                        otherids = aline.get('ids', {})
                        affline += [otherids.get(idname, None) for idname in ['grid', 'wikipedia']]
                        geoinfo = aline.get('geo', {})
                        affline += [geoinfo.get(idname, None) for idname in ['country_code', 'city', 'region']]
                        affline += [load_float(geoinfo.get(idname, None)) for idname in ['latitude', 'longitude']]
                        affline += [load_int(aline.get(idname, None)) for idname in ['works_count', 'cited_by_count']]
                        affiliation_info.append(affline)
                    

            aff = pd.DataFrame(affiliation_info, columns = affil_column_names)
            if preprocess:
                fname = os.path.join(self.path2database, self.path2affiliation, '{}{}.{}'.format(self.path2affiliation, ifile, self.database_extension))
                self.save_data_file(aff, fname, key =self.path2affiliation)

                ifile += 1
                affiliation_info = []

        return aff

    def parse_authors(self, preprocess = True, specific_update='', process_name = True, show_progress=True):
        """
        Parse the OpenAlex Author raw data.

        Parameters
        ----------
        preprocess: bool, default True
            Save the processed data in new DataFrames.

        specific_update: str
            Parse only a specific update date, specified by the date in Y-M-D format, 2022-01-01.  
            If empty the full data is parsed.

        process_name: bool, default True
            If True, then when processing the raw file, the package `NameParser <https://nameparser.readthedocs.io/en/latest/>`_
            will be used to split author FullNames.

        show_progress: bool, default True
            Show progress with processing of the data.

        Returns
        ----------
        DataFrame
            Author DataFrame.
        """

        if specific_update == '' or specific_update is None:
            author_dir = os.path.join(self.path2database, 'data', 'authors')
        else:
            author_dir = os.path.join(self.path2database, 'data/authors', 'updated_date={}'.format(specific_update))

        if not os.path.exists(author_dir):
            raise ValueError("The authors data was not found in the DataBase path.  Please download the data before parsing.")
        else:
            files_to_parse = [os.path.join(dirpath, file) for (dirpath, dirnames, filenames) in os.walk(author_dir) for file in filenames if '.gz' in file]


        author_column_names = ['AuthorId', 'FullName', 'LastKnownAffiliationId', 'ORCID', 'WikiPage', 'NumberPublications', 'NumberCitations']
        if process_name:
            author_column_names += ['LastName', 'FirstName', 'MiddleName']


        if preprocess:
            if not os.path.exists(os.path.join(self.path2database, self.path2author)):
                os.mkdir(os.path.join(self.path2database, self.path2author))

        ifile = 0
        authorinfo = []
        author = pd.DataFrame(authorinfo, columns = author_column_names)
        for file_name in tqdm(files_to_parse, desc='Authors', leave=True, disable=not show_progress):
            with gzip.open(os.path.join(author_dir, file_name), 'r') as infile:
                for line in infile:
                    if line != '\n'.encode():
                        aline = json.loads(line)
                        authline = [self.clean_openalex_ids(aline['id']), aline.get('display_name', None), self.clean_openalex_ids(aline.get('last_known_institution', None))]
                        otherids = aline.get('ids', {})
                        authline += [otherids.get(idname, None) for idname in ['orcid', 'wikipedia']]
                        authline += [load_int(aline.get(idname, None)) for idname in ['works_count', 'cited_by_count']]

                        # process the first, middle, and last names for the author
                        if process_name:
                            name = aline.get('display_name', '')
                            if name is None:
                                name = ''
                            hname = HumanName(unicodedata.normalize('NFD', name))
                            authline += [hname.last, hname.first, hname.middle]

                        authorinfo.append(authline)

                author = pd.DataFrame(authorinfo, columns = author_column_names)

                # time to save
                if preprocess:
                    fname = os.path.join(self.path2database, self.path2author, '{}{}.{}'.format(self.path2author, ifile, self.database_extension))
                    self.save_data_file(author, fname, key =self.path2author)

                    ifile += 1
                    authorinfo = []

        return author

    def parse_venues(self, preprocess = True, specific_update='', show_progress=True):
        """
        Parse the OpenAlex Venues raw data.

        Parameters
        ----------
        preprocess: bool, default True
            Save the processed data in new DataFrames.

        specific_update: str
            Parse only a specific update date, specified by the date in Y-M-D format, 2022-01-01.  
            If empty the full data is parsed.

        show_progress: bool, default True
            Show progress with processing of the data.

        Returns
        ----------
        DataFrame
            Venue DataFrame.
        """
        if specific_update == '' or specific_update is None:
            venue_dir = os.path.join(self.path2database, 'data', 'venues')
        else:
            venue_dir = os.path.join(self.path2database, 'data/venues', 'updated_date={}'.format(specific_update))

        if not os.path.exists(venue_dir):
            raise ValueError("The venues data was not found in the DataBase path.  Please download the data before parsing.")
        else:
            files_to_parse = [os.path.join(dirpath, file) for (dirpath, dirnames, filenames) in os.walk(venue_dir) for file in filenames if '.gz' in file]


        if preprocess:
            if not os.path.exists(os.path.join(self.path2database, self.path2journal)):
                os.mkdir(os.path.join(self.path2database, self.path2journal))


        venue_column_names = ['VenueId', 'FullName', 'ISSN', 'ISSN_l', 'HomePage', 'NumberPublications', 'NumberCitations']
        venue_info = []
        ifile = 0
        for file_name in tqdm(files_to_parse, desc='Venues', leave=True, disable=not show_progress):
            with gzip.open(os.path.join(venue_dir, file_name), 'r') as infile:
                for line in infile:
                    if line != '\n'.encode():
                        vline = json.loads(line)
                        vdata = [self.clean_openalex_ids(vline['id'])]
                        vdata += [vline.get(idname, None) for idname in ['display_name', 'issn', 'issn_l', 'homepage_url']]
                        vdata += [load_int(vline.get(idname, None)) for idname in ['works_count', 'cited_by_count']]
                        venue_info.append(vdata)

                venue = pd.DataFrame(venue_info, columns = venue_column_names)
                if preprocess:
                    fname = os.path.join(self.path2database, self.path2journal, '{}{}.{}'.format(self.path2journal, ifile, self.database_extension))
                    self.save_data_file(venue, fname, key =self.path2journal)

                    ifile += 1
                    venue_info = []
        
        return venue


    def parse_publications(self, preprocess = True, specific_update='', preprocess_dicts = True, num_file_lines=10**6,
        dataframe_list = ['publications', 'references', 'publicationauthoraffiliation', 'fields'],
        show_progress=True):
        """
        Parse the OpenAlex Works raw data.

        Parameters
        ----------
        preprocess: bool, default True
            Save the processed data in new DataFrames.

        specific_update: str
            Parse only a specific update date, specified by the date in Y-M-D format, 2022-01-01.  
            If empty the full data is parsed.

        parse_venues: bool, default True
            Also parse the venue information.

        preprocess_dicts: bool, default True
            Save the processed Year and DocType data as dictionaries.

        num_file_lines: int, default 10**6
            The processed data will be saved into smaller DataFrames, each with `num_file_lines` rows.

        dataframe_list: list
            The data types to download and save from OpenAlex.
                'all'
                'publications'
                'references'
                'publicationauthoraffiliation'
                'fields'
                'abstracts'

        show_progress: bool, default True
            Show progress with processing of the data.

        Returns
        ----------
        DataFrame
            Publication DataFrame.
        """

        if specific_update == '' or specific_update is None:
            work_dir = os.path.join(self.path2database, 'data', 'works')
        else:
            work_dir = os.path.join(self.path2database, 'data/works', 'updated_date={}'.format(specific_update))

        if not os.path.exists(work_dir):
            raise ValueError("The works data was not found in the DataBase path.  Please download the data before parsing.")
        else:
            files_to_parse = [os.path.join(dirpath, file) for (dirpath, dirnames, filenames) in os.walk(work_dir) for file in filenames if '.gz' in file]


        if 'all' in dataframe_list:
            dataframe_list = openalex_works_dfset


        pub_column_names = ['PublicationId', 'JournalId', 'Year', 'NumberCitations', 'Title', 'Date', 'DocType', 'Doi', 'PMID', 'Volume', 'Issue', 'FirstPage', 'LastPage', 'IsRetracted', 'IsParaText']

        if preprocess and (('publications' in dataframe_list) or ('works' in dataframe_list)):
            if not os.path.exists(os.path.join(self.path2database, self.path2pub)):
                os.mkdir(os.path.join(self.path2database, self.path2pub))

            pubinfo = []
            pub2year = {}
            pub2doctype = {}

        if preprocess and ('references' in dataframe_list):
            if not os.path.exists(os.path.join(self.path2database, self.path2pub2ref)):
                os.mkdir(os.path.join(self.path2database, self.path2pub2ref))

            pub2ref = []

        paa_column_names = ['PublicationId', 'AuthorId', 'AffiliationId', 'AuthorSequence', 'OrigAuthorName']

        if preprocess and ('publicationauthoraffiliation' in dataframe_list):
            if not os.path.exists(os.path.join(self.path2database, self.path2paa)):
                os.mkdir(os.path.join(self.path2database, self.path2paa))

            paa = []

        pub2field_column_names = ['PublicationId', 'FieldId', 'FieldLevel', 'Score']

        if preprocess and (('fields' in dataframe_list) or ('concepts' in dataframe_list)):
            if not os.path.exists(os.path.join(self.path2database, self.path2pub2field)):
                os.mkdir(os.path.join(self.path2database, self.path2pub2field))

            pub2field = []

        if preprocess and ('abstracts' in dataframe_list):
            if not os.path.exists(os.path.join(self.path2database, self.path2pub2abstract)):
                os.mkdir(os.path.join(self.path2database, self.path2pub2abstract))

            pub2abstract = []

        ifile = 0
        for file_name in tqdm(files_to_parse, desc='Works', leave=True, disable=not show_progress):
            with gzip.open(os.path.join(work_dir, file_name), 'r') as infile:
                
                iline = 0
                
                for line in infile:
                    if line != '\n'.encode():
                        wline = json.loads(line)
                        
                        ownid = self.clean_openalex_ids(wline['id'])

                        if ('publications' in dataframe_list) or ('works' in dataframe_list):
                            wdata = [ownid, self.clean_openalex_ids(wline.get('host_venue', {}).get('id', None))]
                            wdata += [load_int(wline.get(idname, None)) for idname in ['publication_year', 'cited_by_count']]
                            wdata += [wline.get(idname, None) for idname in ['title', 'publication_date', 'type']]
                            doi = wline.get('doi', None)
                            if isinstance(doi, str):
                                doi = doi.replace("https://doi.org/", "")
                            wdata += [doi]
                            pmid = wline.get("ids", {}).get('pmid', None)
                            if isinstance(pmid, str):
                                pmid = pmid.replace("https://pubmed.ncbi.nlm.nih.gov/", "")
                            wdata += [pmid]
                            bibinfo = wline.get('biblio', {})
                            wdata += [bibinfo.get(idname, None) for idname in ['volume', 'issue', 'first_page', 'last_page']]
                            wdata += [load_bool(wline.get(extraid, None)) for extraid in ['is_retracted', 'is_paratext']]

                            if preprocess_dicts:
                                pub2year[ownid] = wdata[2]
                                pub2doctype[ownid] = wdata[7]

                            pubinfo.append(wdata)

                        if ('references' in dataframe_list):
                            pub2ref.extend([[ownid, self.clean_openalex_ids(citedid)] for citedid in wline.get('referenced_works', [])])


                        if ('publicationauthoraffiliation' in dataframe_list):
                            iauthor = 1
                            for authorinfo in wline.get('authorships', []):
                                authorid = self.clean_openalex_ids(authorinfo.get('author', {}).get('id', None))
                                authorname = authorinfo.get('author', {}).get('display_name', '')
                                
                                if (iauthor == 1 and authorinfo.get('author_position', None) != 'first'):
                                    iauthor = None
                                    #print(wline.get('authorships', []))
                                    #raise ValueError('author position')

                                institution_list = authorinfo.get('institutions', [])
                                if len(institution_list) == 0:
                                    institution_list = [None]
                                else:
                                    institution_list = [self.clean_openalex_ids(affinfo.get('id', None)) for affinfo in institution_list]

                                for inid in institution_list:
                                    paa.append([ownid, authorid, inid, iauthor, authorname])

                                if not iauthor is None:
                                    iauthor += 1

                        if ('fields' in dataframe_list) or ('concepts' in dataframe_list):
                            pub_concepts = wline.get('concepts', [])
                            for con_dict in pub_concepts:
                                pub2field.append([ownid, self.clean_openalex_ids(con_dict.get('id', None)), load_int(con_dict.get('level', None)), load_float(con_dict.get('score', None))])
                        

                        if ('abstracts' in dataframe_list) and 'abstract_inverted_index' in wline:
                            pub2abstract.append([ownid, wline['abstract_inverted_index']])

                        iline += 1

                        if iline % num_file_lines == 0:
                            if ('publications' in dataframe_list) or ('works' in dataframe_list):
                                pub = pd.DataFrame(pubinfo, columns = pub_column_names)
                                if preprocess:
                                    fname = os.path.join(self.path2database, self.path2pub, '{}{}.{}'.format(self.path2pub, ifile, self.database_extension))
                                    self.save_data_file(pub, fname, key =self.path2pub)
                                    
                                    pubinfo = []
                            
                            if ('references' in dataframe_list):
                                pub2refdf = pd.DataFrame(pub2ref, columns = ['CitingPublicationId', 'CitedPublicationId'])
                                if preprocess:
                                    fname = os.path.join(self.path2database, self.path2pub2ref, '{}{}.{}'.format(self.path2pub2ref, ifile, self.database_extension))
                                    self.save_data_file(pub2refdf, fname, key =self.path2pub2ref)

                                    pub2ref = []

                            if ('publicationauthoraffiliation' in dataframe_list):
                                paadf = pd.DataFrame(paa, columns = paa_column_names)
                                if preprocess:
                                    fname = os.path.join(self.path2database, self.path2paa, '{}{}.{}'.format(self.path2paa, ifile, self.database_extension))
                                    self.save_data_file(paadf, fname, key =self.path2paa)

                                    paa = []

                            if ('fields' in dataframe_list) or ('concepts' in dataframe_list):
                                pub2fielddf = pd.DataFrame(pub2field, columns = pub2field_column_names)
                                if preprocess:
                                    fname = os.path.join(self.path2database, self.path2pub2field, '{}{}.{}'.format(self.path2pub2field, ifile, self.database_extension))
                                    self.save_data_file(pub2fielddf, fname, key =self.path2pub2field)

                                    pub2field = []

                            if ('abstracts' in dataframe_list):
                                with gzip.open(os.path.join(self.path2database, self.path2pub2abstract, '{}{}.jsonl.gz'.format(self.path2pub2abstract, ifile)), 'wb') as outfile:
                                    for abentry in pub2abstract:
                                        outfile.write((json.dumps({abentry[0]:abentry[1]}) + "\n").encode('utf-8'))

                                pub2abstract = []

                            ifile += 1


                
                if ('publications' in dataframe_list) or ('works' in dataframe_list):
                    pub = pd.DataFrame(pubinfo, columns = pub_column_names)
                    if preprocess:
                        fname = os.path.join(self.path2database, self.path2pub, '{}{}.{}'.format(self.path2pub, ifile, self.database_extension))
                        self.save_data_file(pub, fname, key =self.path2pub)
                        
                        pubinfo = []
                
                if ('references' in dataframe_list):
                    pub2refdf = pd.DataFrame(pub2ref, columns = ['CitingPublicationId', 'CitedPublicationId'])
                    if preprocess:
                        fname = os.path.join(self.path2database, self.path2pub2ref, '{}{}.{}'.format(self.path2pub2ref, ifile, self.database_extension))
                        self.save_data_file(pub2refdf, fname, key =self.path2pub2ref)

                        pub2ref = []

                if ('publicationauthoraffiliation' in dataframe_list):
                    paadf = pd.DataFrame(paa, columns = paa_column_names)
                    if preprocess:
                        fname = os.path.join(self.path2database, self.path2paa, '{}{}.{}'.format(self.path2paa, ifile, self.database_extension))
                        self.save_data_file(paadf, fname, key =self.path2paa)

                        paa = []

                if ('fields' in dataframe_list) or ('concepts' in dataframe_list):
                    pub2fielddf = pd.DataFrame(pub2field, columns = pub2field_column_names)
                    if preprocess:
                        fname = os.path.join(self.path2database, self.path2pub2field, '{}{}.{}'.format(self.path2pub2field, ifile, self.database_extension))
                        self.save_data_file(pub2fielddf, fname, key =self.path2pub2field)

                        pub2field = []

                if ('abstracts' in dataframe_list):
                    with gzip.open(os.path.join(self.path2database, self.path2pub2abstract, '{}{}.gz'.format(self.path2pub2abstract, ifile)), 'w') as outfile:
                        for abentry in pub2abstract:
                            outfile.write((json.dumps({abentry[0]:abentry[1]}) + "\n").encode('utf-8'))

                    pub2abstract = []

                ifile += 1

        if preprocess_dicts and (('publications' in dataframe_list) or ('works' in dataframe_list)) and preprocess_dicts:
            with gzip.open(os.path.join(self.path2database, 'pub2year.json.gz'), 'w') as outfile:
                outfile.write(json.dumps(pub2year).encode('utf8'))

            with gzip.open(os.path.join(self.path2database, 'pub2doctype.json.gz'), 'w') as outfile:
                outfile.write(json.dumps(pub2doctype).encode('utf8'))

        return True

    def parse_concepts(self, preprocess = True, specific_update='', show_progress=True):
        """
        Parse the MAG Paper Field raw data.

        Parameters
        ----------
        preprocess: bool, default True
            Save the processed data in new DataFrames.

        specific_update: str
            Parse only a specific update date, specified by the date in Y-M-D format, 2022-01-01.  
            If empty the full data is parsed.

        show_progress: bool, default True
            Show progress with processing of the data.

        Returns
        ----------
        DataFrame
            FieldInfo DataFrame.
        """

        if specific_update == '' or specific_update is None:
            concept_dir = os.path.join(self.path2database, 'data', 'concepts')
        else:
            concept_dir = os.path.join(self.path2database, 'data/concepts', 'updated_date={}'.format(specific_update))

        if not os.path.exists(concept_dir):
            raise ValueError("The concepts data was not found in the DataBase path.  Please download the data before parsing.")
        else:
            files_to_parse = [os.path.join(dirpath, file) for (dirpath, dirnames, filenames) in os.walk(concept_dir) for file in filenames if '.gz' in file]


        field_column_names = ['FieldId', 'FieldName', 'WikiData', 'FieldLevel', 'NumberPublications', 'NumberCitations']

        if preprocess:
            if not os.path.exists(os.path.join(self.path2database, self.path2fieldinfo )):
                os.mkdir(os.path.join(self.path2database, self.path2fieldinfo ))

        fieldinfo = []
        fieldhierarchy = []
        for file_name in tqdm(files_to_parse, desc='Concepts', leave=True, disable=not show_progress):
            with gzip.open(os.path.join(concept_dir, file_name), 'r') as infile:
                for line in infile:
                    if line != '\n'.encode():
                        cline = json.loads(line)
                        fid = self.clean_openalex_ids(cline.get('id', None))
                        fielddata = [fid]
                        fielddata += [cline.get(prop, None) for prop in ['display_name', 'wikidata']]
                        fielddata += [load_int(cline.get(prop, None)) for prop in ['level', 'works_count', 'cited_by_count']]
                        
                        fieldinfo.append(fielddata)

                        for an in cline.get('ancestors', []):
                            fieldhierarchy.append([self.clean_openalex_ids(an.get('id', None)), fid])
                
                    

        fieldinfo = pd.DataFrame(fieldinfo, columns = field_column_names)
        if preprocess:
            fname = os.path.join(self.path2database, self.path2fieldinfo, '{}{}.{}'.format(self.path2fieldinfo, 0, self.database_extension))
            self.save_data_file(fieldinfo, fname, key =self.path2fieldinfo)

            fieldhierarchy = pd.DataFrame(fieldhierarchy, columns = ['ParentFieldId', 'ChildFieldId'])
            fname = os.path.join(self.path2database, self.path2fieldinfo, '{}{}.{}'.format('fieldhierarchy', 0, self.database_extension))
            self.save_data_file(fieldhierarchy, fname, key ='fieldhierarchy')

        return fieldinfo





# -*- coding: utf-8 -*-
"""
.. module:: qfactor
    :synopsis: Calculate the qfactor.

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import sys

import pandas as pd
import numpy as np

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

from pyscisci.utils import zip2dict


## Q-factor
def qfactor(a):
    """
    This function calculates the Q-factor for an author.  See :cite:`Sinatra2016individual` for details.

    """

    return np.exp(np.mean(np.log(a[a>0])))

def compute_qfactor(df, colgroupby, colcountby, show_progress=False):
    """
    Calculate the q factor for each group in the DataFrame.  See :cite:`Sinatra2016individual` for the definition.

    The algorithmic implementation for each author can be found in :py:func:`citationanalysis.author_qfactor`.

    Parameters
    ----------
    :param df : DataFrame
        A DataFrame with the citation information for each Author.

    :param colgroupby : str
        The DataFrame column with Author Ids.

    :param colcountby : str
        The DataFrame column with Citation counts for each publication.

    Returns
    -------
    DataFrame
        DataFrame with 2 columns: colgroupby, 'Hindex'

        """
    # register our pandas apply with tqdm for a progress bar
    tqdm.pandas(desc='Qfactor', disable= not show_progress)

    newname_dict = zip2dict([str(colcountby), '0'], [str(colgroupby)+'Qfactor']*2)
    return df.groupby(colgroupby, sort=False)[colcountby].progress_apply(qfactor).to_frame().reset_index().rename(columns=newname_dict)


# -*- coding: utf-8 -*-
"""
.. module:: longterm impact
    :synopsis: Set of functions for typical bibliometric citation analysis

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import pandas as pd
import numpy as np

import scipy.optimize as spopt
import scipy.stats as spstats

from pyscisci.utils import zip2dict


def _fit_predicted_citations(publication_citations):

    recenter_time = np.sort(publication_citations['DateDelta'].values)

    def fit_f(x):
        return np.arange(1,len(recenter_time) + 1) - np.array([predicted_c(t, x[0], x[1], x[2]) for t in recenter_time])

    s, _ = spopt.leastsq(fit_f, x0 = np.ones(3))
    return pd.Series(s)

def predicted_c(t, lam, mu, sig, m = 30.):
    lognormt = (np.log(t) - mu) / sig
    return m * (np.exp(lam * spstats.norm.cdf( lognormt ) ) - 1.0)

def longterm_impact(pub2ref, colgroupby = 'CitedPublicationId', coldate='CitingYear', show_progress=True):
    """
    This function calculates the longterm scientific impact as introduced in :cite:`Wang2013longterm`.

    Following equation (3) from [w]:
    c_(t) = m * (e^{lam * PHI()})

    """
    pub2ref = pub2ref.copy()

    if 'Year' in coldate:
        pub2ref['DateDelta'] = pub2ref.groupby(colgroupby, sort=False)[coldate].transform(lambda x: x-x.min())
    elif 'Date' in coldate:
        pub2ref['DateDelta'] = pub2ref.groupby(colgroupby, sort=False)[coldate].transform(lambda x: x-x.min()) / np.timedelta64(1,'D')
    else:
        print("Column Date Error") 

    pub2ref = pub2ref.loc[pub2ref['DateDelta'] > 0]
    pub2ref.sort_values(by=['DateDelta'], inplace=True)

    newname_dict = zip2dict(list(range(4)), ['lam', 'mu', 'sig', 'm' ])
    return pub2ref.groupby(colgroupby, sort=False).apply(_fit_predicted_citations).reset_index().rename(columns = newname_dict)



# -*- coding: utf-8 -*-
"""
.. module:: hotstreak
    :synopsis: Calculate the career hotstreak.

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """
import sys
import pandas as pd
import numpy as np

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

from pyscisci.utils import hard_rolling_window


def piecewise_step(x, a,b,c,d):
    return np.piecewise(x, [x<a, np.logical_and(x>=a,x<=b), x>b], [c,d,c])

def piecewise_step_err(y, loc1, loc2):
    y1 = y[loc1:(loc2+1)]
    y0 = np.hstack([y[:loc1], y[(loc2+1):]])
    return np.sqrt(np.sum([np.sum((a - a.mean())**2) for a in [y0,y1]]))

def brut_fit_piecewise_step(ys):
    locs = np.vstack(np.triu_indices(ys.shape[0]-1, k=1)).T + 1
    errs = [piecewise_step_err(ys, i, j) for i,j in locs]
    
    return np.min(errs), locs[np.argmin(errs)]


def piecewise_step_err2(y, loc1, loc2, loc3, loc4):
    y1 = y[loc1:(loc2+1)]
    y2 = y[loc3:(loc4+1)]
    y0 = np.hstack([y[:loc1], y[(loc2+1):loc3], y[(loc4+1):]])
    return np.sqrt(np.sum([np.sum((a - a.mean())**2) for a in [y0,y1,y2]]))

def brut_fit_piecewise_step2(ys):
    locs1 = np.vstack(np.triu_indices(ys.shape[0]-3, k=1)).T + 1
    locs2 = np.array([[i,j,j+k,j+m] for i,j in locs1 for k,m in np.vstack(np.triu_indices(ys.shape[0]-j-1, k=1)).T])
    errs = [piecewise_step_err2(ys, i, j,k,m) for i,j,k,m in locs2]
    
    return np.min(errs), locs2[np.argmin(errs)]

def career_hotstreak(author_career_df, citecol='c10', maxk=1, l1_lambda = 1.):
    """
    Identify hot streaks in author careers :cite:`liu2018hotstreak'.

    TODO: this is an interger programming problem.  Reimplement using an interger solver.
    Right now just using a brut force search (very inefficient)!
    
    Parameters
    ----------
    author_career_df : DataFrame
        The author publication history.

    citecol : str, default 'c10'
        The column with publication citation information.
    
    max_k : int, default 1
        The maximum number of hot streaks to search for in a career. Should be 1 or 2.
    
    l1_lambda : float, default 1.0
        The l1 regularization for the number of streaks.  
        Note, the authors never define the value they used for this in the SI.
        
    Returns
    ----------
    lsm_err : float
        The least square mean error of the model plus the l1-regularized term for the number of model coefficients.

    streak_loc : array
        The index locations for the hot streak start and end locations.
    """
    if maxk == 0 or maxk > 2:
        raise NotImplementedError("the career hotstreak is not implemented for this number of streaks.  set maxk = 1 or maxk=2 ")

    Delta_N = max(5, int(0.1*author_career_df.shape[0]))
    gamma_N = hard_rolling_window(np.log10(author_career_df[citecol].values), window=Delta_N, step_size = 1).mean(axis=1)
    gamma_N = gamma_N[int(Delta_N/2):-int(Delta_N/2)]
    
    nostreak_err = np.sqrt(np.sum((gamma_N - gamma_N.mean())**2)) + l1_lambda # no step functions uses 1 model coefficient
    streak_gammas = [gamma_N.mean(), None, None]

    streak_err, streak_loc1 = brut_fit_piecewise_step(gamma_N)
    streak_err += 3*l1_lambda # 1 step function = 3 model coefficients
    
    streak_loc = [None]*4

    if (nostreak_err <= streak_err): 
        streak_err = nostreak_err
        nstreak = 0
    else:
        streak_loc[:2] = streak_loc1 + int(Delta_N/2)
        streak_gammas[0] = np.hstack([gamma_N[:streak_loc1[0]], gamma_N[(streak_loc1[1]+1):]]).mean()
        streak_gammas[1] = gamma_N[streak_loc1[0]:(streak_loc1[1]+1)].mean()
        nstreak = 1

    if maxk == 2:
        streak_err2, streak_loc2 = brut_fit_piecewise_step2(gamma_N)
        streak_err2 += 6*l1_lambda # 2 step functions = 6 model coefficients
        if (streak_err > streak_err2): 
            streak_err, streak_loc = streak_err2, list(streak_loc2 + int(Delta_N/2))
            streak_gammas[0] = np.hstack([gamma_N[:streak_loc1[0]], gamma_N[streak_loc1[1]:(streak_loc1[2])], gamma_N[(streak_loc1[3]+1):]]).mean()
            streak_gammas[1] = gamma_N[streak_loc1[0]:(streak_loc1[1]+1)].mean()
            streak_gammas[2] = gamma_N[streak_loc1[2]:(streak_loc1[3]+1)].mean()
            nstreak = 2
    
    solution_df = [[streak_gammas[0], streak_gammas[1], streak_loc[0], streak_loc[1]]]
    if nstreak == 2:
        solution_df += [streak_gammas[0], streak_gammas[2], streak_loc[3], streak_loc[4]]
    return pd.DataFrame(solution_df, columns = ['Baseline', 'StreakGamma', 'StreakStart', 'StreakEnd'])



import os
import sys
import json
import gzip
import zipfile

import pandas as pd
import numpy as np
from nameparser import HumanName

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

from pyscisci.datasource.readwrite import load_preprocessed_data, load_int, load_float, load_html_str
from pyscisci.database import BibDataBase
from pyscisci.utils import download_file_from_google_drive

# hide this annoying performance warnings
import warnings
warnings.simplefilter(action='ignore', category=pd.errors.PerformanceWarning)


class CustomDB(BibDataBase):
    """
    Base class for creating a CustomDB.

    """

    def __init__(self, path2database = '', database_extension='csv.gz', keep_in_memory = False, global_filter=None, 
        enable_dask=False, show_progress=True):

        self._default_init(path2database, database_extension, keep_in_memory, global_filter, enable_dask, show_progress)

        self.PublicationIdType = int
        self.AffiliationIdType = int
        self.AuthorIdType = int
        self.JournalIdType = int

    def set_new_data_paths(new_path_dict={}):
        """
        Override path to the dataframe collections based on a new custom hierarchy.

        Parameters
        --------
        new_path_dict : dict
            A dictionary where each key is a dataframe name to override.  E.g. 'author', 'pub', 'paa', 'pub2field', etc.
            and each item is the new dataframe path.

        """
        for dfname, new_path in new_path_dict.items():
            self.set_new_data_path(dfname, new_path)



import os
import sys
import json
import gzip

import pandas as pd
import numpy as np
from nameparser import HumanName
import requests
from lxml import etree
from io import BytesIO

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

from pyscisci.datasource.readwrite import load_preprocessed_data, load_int, load_float, load_html_str, load_xml_text
from pyscisci.database import BibDataBase


class WOS(BibDataBase):
    """
    Base class for Web of Science interface.

    """

    def __init__(self, path2database = '', database_extension='csv.gz', keep_in_memory = False, global_filter=None, 
        enable_dask=False, show_progress=True):

        self._default_init(path2database, database_extension, keep_in_memory, global_filter, enable_dask, show_progress)

        self.PublicationIdType = str
        self.AffiliationIdType = str
        self.AuthorIdType = str
        self.JournalIdType = str

    def download_from_source(self):

        raise NotImplementedError("The Web of Science (WOS) is proprietary owned by Clarivate Analytics.  Contact their sales team to aquire access to the data.")

    def parse_affiliations(self, preprocess = False, show_progress=False):
        raise NotImplementedError("WOS is stored as a xml archive.  Run preprocess to parse the archive.")

    def parse_authors(self, preprocess = False, process_name = True, num_file_lines = 5*10**6, show_progress=False):
        raise NotImplementedError("WOS does not contain disambiguated author information.")

    def _blank_wos_publication(self, PublicationId = 0):
        record = {}
        record['PublicationId'] = PublicationId
        record['Title'] = ''
        record['Year'] = 0
        record['Volume'] = 0
        record['Issue'] = ''
        record['Pages'] = ''
        record['JournalId'] = ''
        record['TeamSize'] = 0
        record['Date'] = 1
        record['DocType'] = ''
        record['ISSN'] = ''
        record['Doi'] = ''

        return record

    def _blank_wos_author(self, AuthorId = None):
        record = {}
        record['AuthorId'] = AuthorId
        record['FullName'] = ''
        record['FirstName'] = ''
        record['LastName'] = ''
        record['AuthorOrder'] = None
        record['Affiliations'] = ''

        return record

    def _blank_wos_affiliation(self):
        record = {}
        record['FullAddress'] = ''
        record['Organizations'] = ''
        record['SubOrganizations'] = ''
        record['City'] = ''
        record['Country'] = ''
        return record

    def _save_dataframes(self, ifile, publication, pub_column_names, author, author_columns, paa, pub2ref, affiliation, field):

        publication = pd.DataFrame(publication, columns=pub_column_names)
        publication['PublicationId'] = publication['PublicationId']
        publication['Year'] = publication['Year'].astype(int)
        publication['Volume'] = pd.to_numeric(publication['Volume'])
        publication['TeamSize'] = publication['TeamSize'].astype(int)
        fname = os.path.join(self.path2database, 'publication', 'publication{}.'.format(ifile) + self.database_extension)
        self.save_data_file(publication, fname, key ='publication')

        author = pd.DataFrame(author, columns = author_columns)
        author['AuthorId'] = author['AuthorId'].astype(int)
        fname = os.path.join(self.path2database, 'author', 'author{}.'.format(ifile) + self.database_extension)
        self.save_data_file(author, fname, key ='author')

        paa = pd.DataFrame(paa, columns = ['PublicationId', 'AuthorId', 'AuthorSequence', 'OrigAuthorName'])
        fname = os.path.join(self.path2database, 'publicationauthoraffiliation', 'publicationauthoraffiliation{}.'.format(ifile) + self.database_extension)
        self.save_data_file(paa, fname, key ='publicationauthoraffiliation')

        pub2ref = pd.DataFrame(pub2ref, columns = ['CitingPublicationId', 'CitedPubliationId'])
        fname = os.path.join(self.path2database, 'pub2ref', 'pub2ref{}.'.format(ifile) + self.database_extension)
        self.save_data_file(pub2ref, fname, key ='pub2ref')

        affiliation = pd.DataFrame(affiliation, columns = ['PublicationId', 'AffiliationId', 'AffiliationString'])
        fname = os.path.join(self.path2database, 'affiliation', 'affiliation{}.'.format(ifile) + self.database_extension)
        self.save_data_file(affiliation, fname, key ='affiliation')
        
        field = pd.DataFrame(field, columns = ['PublicationId', 'FieldId', 'FieldType'])
        fname = os.path.join(self.path2database, 'pub2field', 'pub2field{}.'.format(ifile) + self.database_extension)
        self.save_data_file(field, fname, key ='pub2field')

    def preprocess(self, xml_directory = 'RawXML', name_space = None,
        process_name=True, num_file_lines=10**6, show_progress=True):
        """
        Bulk preprocess of the Web of Science raw data.

        Parameters
        ----------
        'xml_directory': str, default 'RawXML'
            The subdirectory containing the raw WOS xml files.

        'name_space': str, defulat None
            The link to a xml namespace file.  Originally 'http://scientific.thomsonreuters.com/schema/wok5.4/public/FullRecord' but 
            this link is now broken and Clarivate has not replaced the namespace.

        process_name: bool, default True
            If True, then when processing the raw file, the package `NameParser <https://nameparser.readthedocs.io/en/latest/>`_
            will be used to split author FullNames.

        num_file_lines: int, default 10**6
            The processed data will be saved into smaller DataFrames, each with `num_file_lines` rows.

        show_progress: bool, default True
            Show progress with processing of the data.

        """

        pub_column_names = ['PublicationId', 'Year', 'JournalId', 'Doi', 'ISSN', 'Title', 'Date', 'Volume', 'Issue', 'Pages', 'DocType', 'TeamSize']
        author_column_names = ['AuthorId', 'FullName', 'FirstName', 'LastName']

        if show_progress:
            print("Starting to preprocess the WOS database.")


        #and we need to standardize the document types
        # as of 7/2021: Book, BookChapter, Conference, Dataset, Journal, Patent, Repository, Thesis, NULL : unknown
        doctype = {'Article': 'j', 'Book Review':'br', 'Letter':"l", 'Review':"rev", 'Correction':"corr",
       'Editorial Material':"editorial", 'News Item':"news", 'Bibliography':"bib",
       'Biographical-Item':"bibi", 'Hardware Review':"hr", 'Meeting Abstract':"ma", 'Note':"note",
       'Discussion':"disc", 'Item About an Individual':"indv", 'Correction, Addition':"corradd",
       'Chronology':"chrono", 'Software Review':"sr", 'Reprint':"re", 'Database Review':"dr", 
       "Journal":'j', 'Book':'b', '':'', 'BookChapter':'bc', 'Conference':'c', 'Dataset':'d', 'Patent':'p', 'Repository':'r', 'Thesis':'t',
       'article': 'j', 'book':'b', '':'', 'phdthesis':'t', 'proceedings':'c', 'inproceedings':'c',
        'mastersthesis':'mst', 'incollection':'coll'}


        for hier_dir_type in ['publication', 'author', 'publicationauthoraffiliation', 'pub2field', 'pub2ref', 'affiliation']:

            if not os.path.exists(os.path.join(self.path2database, hier_dir_type)):
                os.mkdir(os.path.join(self.path2database, hier_dir_type))


        pub2year = {}
        pub2doctype = {}

        found_aids = set([])

        found_affiliations = {}


        ns = {"ns": name_space}
        xmlfiles = sorted([fname for fname in os.listdir(os.path.join(self.path2database, xml_directory)) if '.xml' in fname])

        ifile = 0
        for xml_file_name in tqdm(xmlfiles, desc='WOS xml files', leave=True, disable=not show_progress):

            publication = []
            author = []
            paa = []
            pub2field = []
            pub2ref = []
            affiliation = []
            field = []

            name, extension = os.path.splitext(xml_file_name)

            if extension == '.gz':
                with gzip.open(os.path.join(self.path2database, xml_directory, xml_file_name), 'r') as infile:
                    xml_file = infile.read()
                xml_file = BytesIO(xml_file)

            elif extension == '.xml':
                with open(os.path.join(self.path2database, xml_directory, xml_file_name), 'r') as infile:
                    xml_file = infile.read()

            # extract the desired fields from the XML tree  #
            
            xmltree = etree.iterparse(xml_file, events=('end',), tag="{{{0}}}REC".format(name_space))

            #if show_progress:
            #     print("{} Xml tree parsed, iterating through elements.".format(xml_file_name))

            last_position = 0

            for event, elem in xmltree:

                # scrape the publication information
                PublicationId = load_html_str(elem.xpath('./ns:UID', namespaces=ns)[0].text)

                pub_record = self._blank_wos_publication(PublicationId)

                pub_record['Title'] = load_html_str(load_xml_text(elem.xpath('./ns:static_data/ns:summary/ns:titles/ns:title[@type="item"]', namespaces=ns)))
                pub_record['JournalId'] = load_html_str(load_xml_text(elem.xpath('./ns:static_data/ns:summary/ns:titles/ns:title[@type="source"]', namespaces=ns)))

                pub_info = elem.xpath('./ns:static_data/ns:summary/ns:pub_info', namespaces=ns)[0]
                pub_record['Year'] = load_int(pub_info.get('pubyear', ''))
                pub_record['Date'] = load_html_str(pub_info.get('sortdate', ''))
                pub_record['Volume'] = load_int(pub_info.get('vol', ''))
                pub_record['Issue'] = load_int(pub_info.get('issue', ''))

                pub2year[PublicationId] = pub_record['Year']

                pub_record['Pages'] = load_html_str(load_xml_text(elem.xpath('./ns:static_data/ns:summary/ns:pub_info/ns:page', namespaces=ns), default=''))

                for ident in ['ISSN', 'Doi']:
                    identobject = elem.xpath('./ns:dynamic_data/ns:cluster_related/ns:identifiers/ns:identifier[@type="{}"]'.format(ident.lower()), namespaces=ns)
                    if len(identobject) > 0:
                        pub_record[ident] =load_html_str( identobject[0].get('value', ''))


                #load_html_str(load_xml_text(elem.xpath('./ns:dynamic_data/ns:cluster_related/ns:identifiers/ns:identifier[@type="doi"]', namespaces=ns)))

                pub_record['DocType'] = load_html_str(load_xml_text(elem.xpath('./ns:static_data/ns:summary/ns:doctypes/ns:doctype', namespaces=ns)))

                pub2doctype[PublicationId] = pub_record['DocType']

                # now scrape the authors
                pub_authors = {}
                author_objects = elem.xpath('./ns:static_data/ns:summary/ns:names/ns:name[@role="author"]', namespaces=ns)
                pub_record['TeamSize'] = len(author_objects)

                for author_obj in author_objects:
                    author_record = self._blank_wos_author(None)
                    author_record['AuthorId'] = author_obj.get('dais_id', None)

                    author_record['FullName'] = load_html_str(load_xml_text(author_obj.xpath('./ns:full_name', namespaces=ns)))
                    author_record['FirstName'] = load_html_str(load_xml_text(author_obj.xpath('./ns:first_name', namespaces=ns)))
                    author_record['LastName'] = load_html_str(load_xml_text(author_obj.xpath('./ns:last_name', namespaces=ns)))

                    author_record['Affiliations'] = author_obj.get('addr_no', '')
                    author_record['Affiliations'] = [int(single_addr_no) for single_addr_no in author_record['Affiliations'].split(' ') if len(single_addr_no) > 0]

                    author_record['AuthorOrder'] = int(author_obj.get('seq_no', None))

                    pub_authors[author_record['AuthorOrder']] = author_record


                #contributor_objects = elem.xpath('./ns:static_data/ns:contributors/ns:contributor/ns:name[@role="researcher_id"]', namespaces=ns)

                address_objects = elem.xpath('./ns:static_data/ns:fullrecord_metadata/ns:addresses/ns:address_name/ns:address_spec', namespaces=ns)
                for addr_obj in address_objects:
                    addr_record = self._blank_wos_affiliation()

                    organization_objects = addr_obj.xpath('./ns:organizations/ns:organization[@pref="Y"]', namespaces=ns)
                    if len(organization_objects) == 0:
                        organization_objects = addr_obj.xpath('./ns:organizations/ns:organization', namespaces=ns)

                    if len(organization_objects) == 0:
                        orgtext = ''
                    else:
                        orgtext = organization_objects[0].text
                    
                    address_no = int(addr_obj.get('addr_no'))

                    affiliation.append([PublicationId, address_no, orgtext])

                    #if found_affiliations

                    #article['addresses'][address_no] = address_info


                field_objects = elem.xpath('./ns:static_data/ns:fullrecord_metadata/ns:category_info/ns:headings/ns:heading', namespaces=ns)
                field.extend([[PublicationId, field_obj.text, 'heading'] for field_obj in field_objects if field_obj is not None])

                field_objects = elem.xpath('./ns:static_data/ns:fullrecord_metadata/ns:category_info/ns:subheadings/ns:subheading', namespaces=ns)
                field.extend([[PublicationId, field_obj.text, 'subheading'] for field_obj in field_objects if field_obj is not None])

                field_objects = elem.xpath('./ns:static_data/ns:fullrecord_metadata/ns:category_info/ns:subjects/ns:subject[@ascatype="traditional"]', namespaces=ns)
                field.extend([[PublicationId, field_obj.text, 'ASCA traditional subject'] for field_obj in field_objects if field_obj is not None])

                field_objects = elem.xpath('./ns:static_data/ns:fullrecord_metadata/ns:category_info/ns:subjects/ns:subject[@ascatype="extended"]', namespaces=ns)
                field.extend([[PublicationId, field_obj.text, 'ASCA extended subject'] for field_obj in field_objects if field_obj is not None])

                field_objects = elem.xpath('./ns:static_data/ns:fullrecord_metadata/ns:keywords/ns:keyword', namespaces=ns)
                field.extend([[PublicationId, field_obj.text, 'keyword'] for field_obj in field_objects if field_obj is not None])

                field_objects = elem.xpath('./ns:static_data/ns:item/ns:keywords_plus/ns:keyword', namespaces=ns)
                field.extend([[PublicationId, field_obj.text, 'keyword plus'] for field_obj in field_objects if field_obj is not None])

                reference_objects = elem.xpath('./ns:static_data/ns:fullrecord_metadata/ns:references/ns:reference', namespaces=ns)
                for ref_obj in reference_objects:
                    for ref_elem in ref_obj:
                        if ref_elem.tag == "{{{0}}}uid".format(name_space):
                            refid = load_html_str(ref_elem.text.replace('WOS:', ''))
                            pub2ref.append([PublicationId, refid])
                        elif ref_elem.tag == "{{{0}}}year".format(name_space):
                            pub2year[refid] = load_int(ref_elem.text)

                publication.append([pub_record[k] for k in pub_column_names])

                for aorder, author_record in pub_authors.items():
                    if not author_record['AuthorId'] is None and not author_record['AuthorId'] in found_aids:
                        found_aids.add(author_record['AuthorId'])
                        author.append([author_record[k] for k in author_column_names])

                    paa.append([PublicationId, author_record['AuthorId'], aorder, author_record['FullName']])


            self._save_dataframes(ifile, publication, pub_column_names, author, author_column_names, paa, pub2ref, affiliation, field)
            ifile += 1

        with gzip.open(os.path.join(self.path2database, 'pub2year.json.gz'), 'w') as outfile:
            outfile.write(json.dumps(pub2year).encode('utf8'))

        with gzip.open(os.path.join(self.path2database, 'pub2doctype.json.gz'), 'w') as outfile:
            outfile.write(json.dumps(pub2doctype).encode('utf8'))




# -*- coding: utf-8 -*-
"""
.. module:: credit sharing
    :synopsis: Set of functions for calcuating credit share amongst authors.

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import pandas as pd
import numpy as np

from pyscisci.utils import isin_sorted, groupby_count
from pyscisci.network import cocitation_network

def credit_share(focus_pid, pub2ref, pub2author, temporal=False, normed=False, show_progress=False):
    """
    Calculate the credit share for each author of a publication based on :cite:`Shen2014credit`.

    Parameters
    ----------
    :param focus_pid : int, str
        The focus publication id.

    :param pub2ref : DataFrame
        A DataFrame with the citation information for each Publication.

    :param pub2author : DataFrame
        A DataFrame with the author information for each Publication.

    :param temporal : bool, default False
        If True, compute the adjacency matrix using only publications for each year.

    :param normed : bool, default False
        Normalize the sum of credit share to 1.0

    :param show_progress : bool, default False
        If True, show a progress bar tracking the calculation.

    Returns
    -------
    credit_share, numpy array
        If temporal == False:
            The adjacency matrix for the co-citation network

        If temporal == True:
            A dictionary with key for each year, and value of the adjacency matrix for the cocitation network induced
            by citing publications in that year.

    author2int, dict
        A mapping of the AuthorIds from the focus publication to the column of the credit share vector or matrix (see above).

    """

    # the focus publication's authors
    focus_authors = np.sort(pub2author.loc[pub2author['PublicationId']==focus_pid]['AuthorId'].unique())
    author2int = {aid:i for i, aid in enumerate(focus_authors)}

    if focus_authors.shape[0] > 1:
        # start by getting the co-citation network around the focus publication
        adj_mat, cited2int = cocitation_network(pub2ref, focus_pub_ids=np.sort([focus_pid]), focus_constraint='egocited',
                temporal=temporal, show_progress=show_progress)

        # get the authorships for the publications in the cocitation network
        cocited_pubs = np.sort(list(cited2int.keys()))
        pa = pub2author.loc[isin_sorted(pub2author['PublicationId'].values, cocited_pubs)]

        if cocited_pubs.shape[0] > 0:
            # the credit allocation matrix has a row for each focus author, and a column for each cocited publication (including the focus pub)
            credit_allocation_mat = np.zeros((focus_authors.shape[0], cocited_pubs.shape[0]), dtype = float)

            # for each cocited publication, we count the number of authors
            # and assign to each focus author, their fractional share of the credit (1 divided by the number of authors)
            for cocitedid, adf in pa.groupby('PublicationId'):
                author2row = [author2int[aid] for aid in adf['AuthorId'].unique() if not author2int.get(aid, None) is None]
                if len(author2row) > 0:
                    credit_allocation_mat[author2row, cited2int[cocitedid]] = 1.0/adf['AuthorId'].nunique()

            if temporal:
                # temporal credit allocation - broken down by year

                # we need the temporal citations to the focus article
                focus_citations = groupby_count(pub2ref.loc[isin_sorted(pub2ref['CitedPublicationId'].values, np.sort([focus_pid]))],
                    colgroupby='CitingYear', colcountby='CitingPublicationId', count_unique=True, show_progress=False)
                focus_citations={y:c for y,c in focus_citations[['CitingYear', 'CitingPublicationIdCount']].values}

                # when temporal is True, a temporal adj mat is returned where each key is the year
                years = np.sort(list(adj_mat.keys()))

                cocite_counts = np.zeros((years.shape[0], cocited_pubs.shape[0]), dtype=float)

                for iy, y in enumerate(years):
                    cocite_counts[iy] = adj_mat[y].tocsr()[cited2int[focus_pid]].todense()#set the off-diagonal to be the total co-citations from that year
                    cocite_counts[iy, cited2int[focus_pid]] = focus_citations[y]          #set the diagonal to be the total citations from that year

                cocite_counts = cocite_counts.cumsum(axis=0)

            else:
                # just do credit allocation with the full cocitation matrix
                cocite_counts = adj_mat.tocsr()[cited2int[focus_pid]].todense()

                # the co-citation matrix misses the number of citations to the focus publication
                # so explicitly calculate the number of citations to the focus publication
                cocite_counts[0,cited2int[focus_pid]] = pub2ref.loc[isin_sorted(pub2ref['CitedPublicationId'].values, np.sort([focus_pid]))]['CitingPublicationId'].nunique()

            # credit share is the matrix product of the credit_allocation_mat with cocite_counts
            credit_share = np.squeeze(np.asarray(credit_allocation_mat.dot(cocite_counts.T)))

            # normalize the credit share vector to sum to 1
            if normed:
                credit_share = credit_share/credit_share.sum(axis=0)

            if temporal:
                return credit_share, author2int, years
            else:
                return credit_share, author2int
        else:
            if temporal:
                years = np.sort(pub2ref.loc[pub2ref['CitedPublicationId'] == focus_pid]['CitingYear'].unique())
                return np.array([[None for y in years] for a in author2int]), author2int, years
            else:
                return np.array([None for a in author2int]), author2int

    elif focus_authors.shape[0] == 1:
        if temporal:
            years = np.sort(pub2ref.loc[pub2ref['CitedPublicationId'] == focus_pid]['CitingYear'].unique())
            return np.ones(shape=(1,years.shape[0])), author2int, years
        else:
            return np.array([1.0]), author2int


#!/usr/bin/env python
# encoding: utf-8


from setuptools import setup, find_packages
from pyscisci import __package__, __description__, __version__


setup(name=__package__,
      version=__version__,
      description='Science of Science',
      long_description=__description__,
      classifiers=[
            'Development Status :: 4 - Beta',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3.7',
            'Intended Audience :: Science/Research',
            'Topic :: Scientific/Engineering :: Information Analysis',
      ],
      keywords=["science of science", "citation network", 'bibliometric'],
      url="https://github.com/ajgates42/pyscisci",
      author = 'Alex Gates <ajgates42@gmail.com>',
      license="MIT",
      packages = find_packages(),
      install_requires=[
            'pandas',
            'numpy',
            'scipy',
            'scikit-learn',
            'nameparser',
            'lxml',
            'requests',
            'unidecode',
            'tqdm',
            'dask',
            'numba'
            ],
      extras_require = {
        'nlp':  ["sparse_dot_topn", "python-Levenshtein"],
        'hdf':['tables']},
      include_package_data=True,
      zip_safe=False
      )



__package__ = 'pyscisci'
__title__ = 'pyscisci: A python package for the science of science'
__description__ = 'Lets study science!'

__copyright__ = '2021, Gates, A.J.'

__author__ = """\n""".join([
    'Alexander J Gates <ajgates42@gmail.com>'
])

__version__ = '0.8.11'
__release__ = '0.8.11'




import os
import sys
import json
import gzip

import pandas as pd
import numpy as np
from nameparser import HumanName
import unicodedata

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

from pyscisci.datasource.readwrite import load_preprocessed_data, load_int, load_float
from pyscisci.database import BibDataBase



class MAG(BibDataBase):
    """
    Base class for Microsoft Academic Graph interface.

    This is an extension of 'BibDataBase' with processing functions specific to the MAG.
    See 'BibDataBase' in database.py for details of non-MAG specific functions.

    The MAG comes structured into three folders: mag, advanced, nlp.
    Explain downloading etc.

    """

    def __init__(self, path2database = '', database_extension='csv.gz', keep_in_memory = False, global_filter = None, 
        enable_dask=False, show_progress=True):

        self._default_init(path2database, database_extension, keep_in_memory, global_filter, enable_dask, show_progress)

        self.PublicationIdType = int
        self.AffiliationIdType = int
        self.AuthorIdType = int
        self.JournalIdType = int

    def preprocess(self, dflist = None, combine_conferences =True, show_progress=True):
        """
        Bulk preprocess the MAG raw data.

        Parameters
        ----------
        dflist: list, default None
            The list of DataFrames to preprocess.  If None, all MAG DataFrames are preprocessed.
        
        combine_conferences: bool, default True,
            combine the conference series into the Journals
            
        show_progress: bool, default True
            Show progress with processing of the data.

        """
        if dflist is None:
            dflist = ['affiliation', 'author', 'publication', 'reference', 'publicationauthoraffiliation', 'fields']

        if 'affiliation' in dflist:
            self.parse_affiliations(preprocess = True, show_progress=show_progress)

        if 'author' in dflist:
            self.parse_authors(preprocess = True, show_progress=show_progress)

        if 'publication' in dflist:
            self.parse_publications(preprocess = True, show_progress=show_progress)

        if 'reference' in dflist:
            self.parse_references(preprocess = True, show_progress=show_progress)

        if 'publicationauthoraffiliation' in dflist:
            self.parse_publicationauthoraffiliation(preprocess = True, show_progress=show_progress)

        if 'fields' in dflist:
            self.parse_fields(preprocess=True, show_progress=show_progress)

    def download_from_source(self):
        #TODO: Error Raising when database isnt defined
        raise NotImplementedError("ToDo")

    def parse_affiliations(self, preprocess=True, show_progress=True):
        """
        Parse the MAG Affilation raw data.

        Parameters
        ----------
        preprocess: bool, default True
            Save the processed data in new DataFrames.

        show_progress: bool, default True
            Show progress with processing of the data.


        Returns
        ----------
        DataFrame
            Affiliation DataFrame.
        """

        affil_int_columns = [0, 7, 8]
        affil_str_columns = [3, 4, 5, 6]
        affil_float_columns = [10, 11]

        affil_column_names = ['AffiliationId', 'NumberPublications', 'NumberCitations', 'FullName', 'GridId', 'OfficialPage', 'WikiPage', 'Latitude', 'Longitude']

        file_name = os.path.join(self.path2database, 'mag', 'Affiliations.txt')

        affiliation_info = []
        with tqdm(total=os.path.getsize(file_name), unit='iB', unit_scale=True, desc='Affiliations', leave=True, disable=not show_progress) as pbar:
            with open(file_name, 'r') as infile:
                for line in infile:
                    sline = line.replace('\n', '').split('\t')
                    affline = [load_int(sline[i]) for i in affil_int_columns]
                    affline += [sline[i] for i in affil_str_columns]
                    affline += [load_float(sline[i]) for i in affil_float_columns]
                    affiliation_info.append(affline)

                    # update progress bar
                    pbar.update(sys.getsizeof(line))

        aff = pd.DataFrame(affiliation_info, columns = affil_column_names)

        if preprocess:
            if not os.path.exists(os.path.join(self.path2database, self.path2affiliation)):
                os.mkdir(os.path.join(self.path2database, self.path2affiliation))
                
            fname = os.path.join(self.path2database, self.path2affiliation, '{}{}.{}'.format(self.path2affiliation, 0, self.database_extension))
            self.save_data_file(aff, fname, key =self.path2affiliation)

        return aff

    def parse_authors(self, preprocess = False, process_name = True, num_file_lines = 5*10**6, show_progress=True):
        """
        Parse the MAG Author raw data.

        Parameters
        ----------
        preprocess: bool, default True
            Save the processed data in new DataFrames.

        process_name: bool, default True
            If True, then when processing the raw file, the package `NameParser <https://nameparser.readthedocs.io/en/latest/>`_
            will be used to split author FullNames.

        num_file_lines: int, default 5*10**6
            The processed data will be saved into smaller DataFrames, each with `num_file_lines` rows.

        show_progress: bool, default True
            Show progress with processing of the data.

        Returns
        ----------
        DataFrame
            Author DataFrame.
        """

        author_int_columns = [0, 4, 5, 7]

        author_column_names = ['AuthorId', 'LastKnownAffiliationId', 'NumberPublications', 'NumberCitations', 'FullName']
        if process_name:
            author_column_names += ['LastName', 'FirstName', 'MiddleName']


        if preprocess:
            if not os.path.exists(os.path.join(self.path2database, self.path2author)):
                os.mkdir(os.path.join(self.path2database, self.path2author))

        file_name = os.path.join(self.path2database, 'mag', 'Authors.txt')

        iauthor = 0
        ifile = 0
        authorinfo = []
        with tqdm(total=os.path.getsize(file_name), unit='iB', unit_scale=True, desc='Authors', disable=not show_progress, leave=True) as pbar:
            with open(file_name, 'r') as infile:

                for line in infile:

                    # split the line and keep only the relevant columns
                    sline = line.split('\t')
                    adata = [load_int(sline[ip]) for ip in author_int_columns] + [sline[2]]

                    # process the first, middle, and last names for the author
                    if process_name:
                        hname = HumanName(unicodedata.normalize('NFD', sline[2]))
                        adata += [hname.last, hname.first, hname.middle]

                    authorinfo.append(adata)
                    iauthor += 1

                    # update progress bar
                    pbar.update(sys.getsizeof(line))

                    # time to save
                    if preprocess and iauthor % num_file_lines == 0:
                        fname = os.path.join(self.path2database, self.path2author, '{}{}.{}'.format(self.path2author, ifile, self.database_extension))
                        self.save_data_file(pd.DataFrame(authorinfo, columns = author_column_names), fname, key =self.path2author)

                        ifile += 1
                        authorinfo = []

                author = pd.DataFrame(authorinfo, columns = author_column_names)
                if preprocess:
                    fname = os.path.join(self.path2database, self.path2author, '{}{}.{}'.format(self.path2author, ifile, self.database_extension))
                    self.save_data_file(author, fname, key =self.path2author)

        return author

    def parse_publications(self, preprocess = True, num_file_lines=2*10**6, preprocess_dicts = True, show_progress=True):
        """
        Parse the MAG Publication and Journal raw data.

        Parameters
        ----------
        preprocess: bool, default True
            Save the processed data in new DataFrames.

        num_file_lines: int, default 5*10**6
            The processed data will be saved into smaller DataFrames, each with `num_file_lines` rows.

        preprocess_dicts: bool, default True
            Save the processed Year and DocType data as dictionaries.

        show_progress: bool, default True
            Show progress with processing of the data.

        Returns
        ----------
        DataFrame
            Publication DataFrame.
        """

        # first do the journal information
        journal_str_col = [2, 4, 5, 6]
        journal_column_names = ['JournalId', 'FullName', 'Issn', 'Publisher', 'Webpage']

        if preprocess:
            if not os.path.exists(os.path.join(self.path2database, self.path2journal)):
                os.mkdir(os.path.join(self.path2database, self.path2journal))

        file_name = os.path.join(self.path2database, 'mag', 'Journals.txt')

        journal_info = []
        with tqdm(total=os.path.getsize(file_name), unit='iB', unit_scale=True, desc='Journals', leave=True, disable=not show_progress) as pbar:
            with open(os.path.join(self.path2database, 'mag', 'Journals.txt'), 'r') as infile:
                for line in infile:
                    # split the line and keep only the relevant columns
                    sline = line.replace('\n', '').split('\t')
                    jline = [load_int(sline[0])] + [sline[i] for i in journal_str_col]
                    journal_info.append(jline)

                    # update progress bar
                    pbar.update(sys.getsizeof(line))

        journal = pd.DataFrame(journal_info, columns = journal_column_names)
        if preprocess:
            fname = os.path.join(self.path2database, self.path2journal, '{}{}.{}'.format(self.path2journal, 0, self.database_extension))
            self.save_data_file(journal, fname, key =self.path2journal)

        #now lets do the publication information
        # we need to standardize the document types across the different databases
        # as of 7/2021: Book, BookChapter, Conference, Dataset, Journal, Patent, Repository, Thesis, NULL : unknown
        doctype = {'Article': 'j', 'Book Review':'br', 'Letter':"l", 'Review':"rev", 'Correction':"corr",
       'Editorial Material':"editorial", 'News Item':"news", 'Bibliography':"bib",
       'Biographical-Item':"bibi", 'Hardware Review':"hr", 'Meeting Abstract':"ma", 'Note':"note",
       'Discussion':"disc", 'Item About an Individual':"indv", 'Correction, Addition':"corradd",
       'Chronology':"chrono", 'Software Review':"sr", 'Reprint':"re", 'Database Review':"dr", 
       "Journal":'j', 'Book':'b', '':'', 'BookChapter':'bc', 'Conference':'c', 'Dataset':'d', 'Patent':'p', 'Repository':'r', 'Thesis':'t',
       'article': 'j', 'book':'b', 'phdthesis':'t', 'proceedings':'c', 'inproceedings':'c',
        'mastersthesis':'mst', 'incollection':'coll'}

        pub_int_columns = [0, 7, 11, 22]
        pub_str_columns = [2, 4, 8, 14, 15, 16, 17, 24]
        pub_column_names = ['PublicationId', 'Year', 'JournalId', 'FamilyId',  'Doi', 'Title', 'Date', 'Volume', 'Issue', 'FirstPage', 'LastPage', 'DocSubTypes', 'DocType']

        if preprocess:
            if not os.path.exists(os.path.join(self.path2database, self.path2pub)):
                os.mkdir(os.path.join(self.path2database, self.path2pub))

        file_name = os.path.join(self.path2database, 'mag', 'Papers.txt')

        ipub = 0
        ifile = 0
        pubinfo = []

        pub2year = {}
        pub2doctype = {}
        with tqdm(total=os.path.getsize(file_name), unit='iB', unit_scale=True, desc='Publications', leave=True, disable=not show_progress) as pbar:
            with open(file_name, 'r') as infile:
                for line in infile:
                    # split the line and keep only the relevant columns
                    sline = line.replace('\n', '').split('\t')
                    pline = [load_int(sline[ip]) for ip in pub_int_columns] + [sline[ip] for ip in pub_str_columns] + [doctype[sline[3]]]
                    pub2year[pline[0]] = pline[1]
                    if doctype[sline[3]] != '':
                        pub2doctype[pline[0]] = doctype[sline[3]]

                    pubinfo.append(pline)
                    ipub += 1

                    # update progress bar
                    pbar.update(sys.getsizeof(line))

                    if preprocess and ipub % num_file_lines == 0:
                        fname = os.path.join(self.path2database, self.path2pub, '{}{}.{}'.format(self.path2pub, ifile, self.database_extension))
                        self.save_data_file(pd.DataFrame(pubinfo, columns = pub_column_names), fname, key =self.path2pub)

                        ifile += 1
                        pubinfo = []

            pub = pd.DataFrame(pubinfo, columns = pub_column_names)
            if preprocess:
                fname = os.path.join(self.path2database, self.path2pub, '{}{}.{}'.format(self.path2pub, ifile, self.database_extension))
                self.save_data_file(pub, fname, key =self.path2pub)

                if preprocess_dicts:
                    with gzip.open(os.path.join(self.path2database, 'pub2year.json.gz'), 'w') as outfile:
                        outfile.write(json.dumps(pub2year).encode('utf8'))

                    with gzip.open(os.path.join(self.path2database, 'pub2doctype.json.gz'), 'w') as outfile:
                        outfile.write(json.dumps(pub2doctype).encode('utf8'))

        return pub


    def parse_references(self, preprocess = False, num_file_lines=10**7, show_progress=True):
        """
        Parse the MAG References raw data.

        Parameters
        ----------
        preprocess: bool, default True
            Save the processed data in new DataFrames.

        num_file_lines: int, default 10**7
            The processed data will be saved into smaller DataFrames, each with `num_file_lines` rows.

        show_progress: bool, default True
            Show progress with processing of the data.

        Returns
        ----------
        DataFrame
            Pub2Ref DataFrame.
        """
        if preprocess:
            if not os.path.exists(os.path.join(self.path2database, self.path2pub2ref)):
                os.mkdir(os.path.join(self.path2database, self.path2pub2ref))

        file_name = os.path.join(self.path2database, 'mag', 'PaperReferences.txt')

        iref = 0
        ifile = 0
        pub2ref_info = []
        with tqdm(total=os.path.getsize(file_name), unit='iB', unit_scale=True, unit_divisor=1024, desc='References', leave=True, disable=not show_progress) as pbar:
            with open(file_name, 'r') as infile:
                for line in infile:
                    # split the line and keep only the relevant columns
                    sline = line.replace('\n', '').split('\t')
                    pub2ref_info.append([load_int(sline[ip]) for ip in range(2)])
                    iref += 1

                    # update progress bar
                    pbar.update(sys.getsizeof(line))

                    if preprocess and iref % num_file_lines == 0:
                        fname = os.path.join(self.path2database, self.path2pub2ref, '{}{}.{}'.format(self.path2pub2ref, ifile, self.database_extension))
                        self.save_data_file(pd.DataFrame(pub2ref_info, columns = ['CitingPublicationId', 'CitedPublicationId']), 
                            fname, key =self.path2pub2ref)

                        ifile += 1
                        pub2ref_info = []

        pub2ref = pd.DataFrame(pub2ref_info, columns = ['CitingPublicationId', 'CitedPublicationId'])

        if preprocess:
            fname = os.path.join(self.path2database, self.path2pub2ref, '{}{}.{}'.format(self.path2pub2ref, ifile, self.database_extension))
            self.save_data_file(pub2ref, fname, key =self.path2pub2ref)

        return pub2ref

    def parse_publicationauthoraffiliation(self, preprocess = False, num_file_lines=5*10**6, show_progress=True):
        """
        Parse the MAG PublicationAuthorAffiliation raw data.

        Parameters
        ----------
        preprocess: bool, default True
            Save the processed data in new DataFrames.

        num_file_lines: int, default 10**7
            The processed data will be saved into smaller DataFrames, each with `num_file_lines` rows.

        show_progress: bool, default True
            Show progress with processing of the data.

        Returns
        ----------
        DataFrame
            PublicationAuthorAffiliation DataFrame.
        """
        pubauthaff_int_columns = [0, 1, 2, 3]
        pubauthaff_str_columns = [4, 5]
        pub_column_names = ['PublicationId', 'AuthorId', 'AffiliationId', 'AuthorSequence',  'OrigAuthorName', 'OrigAffiliationName']

        if preprocess:
            if not os.path.exists(os.path.join(self.path2database, self.path2paa)):
                os.mkdir(os.path.join(self.path2database, self.path2paa))

        file_name = os.path.join(self.path2database, 'mag', 'PaperAuthorAffiliations.txt')

        iref = 0
        ifile = 0
        pubauthaff_info = []
        with tqdm(total=os.path.getsize(file_name), unit='iB', unit_scale=True, desc='PaperAuthorAffiliations', leave=True, disable=not show_progress) as pbar:
            with open(file_name, 'r') as infile:
                for line in infile:
                    sline = line.replace('\n', '').split('\t')
                    pubauthaff_info.append([load_int(sline[ip]) for ip in pubauthaff_int_columns] + [sline[ip] if len(sline) > ip else '' for ip in pubauthaff_str_columns ])
                    iref += 1

                    # update progress bar
                    pbar.update(sys.getsizeof(line))

                    if preprocess and iref % num_file_lines == 0:

                        fname = os.path.join(self.path2database, self.path2paa, '{}{}.{}'.format(self.path2paa, ifile, self.database_extension))
                        self.save_data_file(pd.DataFrame(pubauthaff_info, columns = pub_column_names), fname, key =self.path2paa)

                        ifile += 1
                        pubauthaff_info = []


        paa = pd.DataFrame(pubauthaff_info, columns = pub_column_names)
        if preprocess:
            fname = os.path.join(self.path2database, self.path2paa, '{}{}.{}'.format(self.path2paa, ifile, self.database_extension))
            self.save_data_file(paa, fname, key =self.path2paa)

        return paa

    def parse_fields(self, preprocess = False, num_file_lines=5*10**6, show_progress=True):
        """
        Parse the MAG Paper Field raw data.

        Parameters
        ----------
        preprocess: bool, default True
            Save the processed data in new DataFrames.

        num_file_lines: int, default 10**7
            The processed data will be saved into smaller DataFrames, each with `num_file_lines` rows.

        show_progress: bool, default True
            Show progress with processing of the data.

        Returns
        ----------
        DataFrame
            Pub2Field DataFrame.
        """
        field2get = [0, 5, 6]
        fieldnames = ['FieldId', 'FieldLevel', 'NumberPublications', 'FieldName']

        if preprocess:
            if not os.path.exists(os.path.join(self.path2database, 'fieldinfo')):
                os.mkdir(os.path.join(self.path2database, 'fieldinfo'))

        fieldinfo = []
        with open(os.path.join(self.path2database, 'advanced', 'FieldsOfStudy.txt'), 'r') as infile:

            for line in infile:
                sline = line.split('\t')
                fielddata = [load_int(sline[ip]) for ip in field2get] + [sline[2]]
                fieldinfo.append(fielddata)

        field = pd.DataFrame(fieldinfo, columns = fieldnames)
        if preprocess:
            fname = os.path.join(self.path2database, self.path2fieldinfo, '{}{}.{}'.format(self.path2fieldinfo, 0, self.database_extension))
            self.save_data_file(field, fname, key =self.path2fieldinfo)


        # and now do pub2field
        paperfields = [0, 1]
        paperfieldnames = ['PublicationId', 'FieldId']

        if preprocess:
            if not os.path.exists(os.path.join(self.path2database, self.path2pub2field)):
                os.mkdir(os.path.join(self.path2database, self.path2pub2field))

        file_name = os.path.join(self.path2database, 'advanced', 'PaperFieldsOfStudy.txt')

        ipaper = 0
        ifile = 0
        fieldinfo = []
        with tqdm(total=os.path.getsize(file_name), unit='iB', unit_scale=True, desc='PaperFieldsOfStudy', leave=True, disable=not show_progress) as pbar:
            with open(file_name, 'r') as infile:

                for line in infile:
                    sline = line.split('\t')
                    fielddata = [int(sline[ip]) for ip in paperfields]
                    fieldinfo.append(fielddata)
                    ipaper += 1

                    # update progress bar
                    pbar.update(sys.getsizeof(line))

                    if preprocess and ipaper % num_file_lines == 0:
                        fname = os.path.join(self.path2database, self.path2pub2field, '{}{}.{}'.format(self.path2pub2field, ifile, self.database_extension))
                        self.save_data_file(pd.DataFrame(fieldinfo, columns = paperfieldnames), fname, key =self.path2pub2field)

                        ifile += 1
                        fieldinfo = []

        pub2field = pd.DataFrame(fieldinfo, columns = paperfieldnames)
        if preprocess:
            fname = os.path.join(self.path2database, self.path2pub2field, '{}{}.{}'.format(self.path2pub2field, ifile, self.database_extension))
            self.save_data_file(pub2field, fname, key =self.path2pub2field)

        return pub2field


